""" Frontend of the project (Argument handling and stuff) """

import sys
from pathlib import Path
from argparse import ArgumentParser
from maketree.core.parser import Parser, ParseError
from maketree.core.tree_builder import TreeBuilder
from maketree.core.normalizer import Normalizer
from maketree.utils import get_existing_paths, print_on_true, print_tree
from typing import List, Dict, Tuple


PROGRAM = "maketree"
VERSION = "1.0.1"


def main():
    args = parse_args()

    sourcefile = Path(args.src)
    dstpath = Path(args.dst)
    VERBOSE: bool = args.verbose
    OVERWRITE: bool = args.overwrite
    SKIP: bool = args.skip
    PRINT_TREE = args.graphical

    # SRC Exists?
    if not sourcefile.exists():
        error("source '%s' does not exist." % sourcefile)

    # SRC Tree file?
    if not sourcefile.name.endswith(".tree"):
        error("source '%s' is not a .tree file." % sourcefile)

    # DST Exists?
    if not dstpath.exists():
        error("destination path '%s' does not exist." % dstpath)

    # DST not a Dir?
    if not dstpath.is_dir():
        error("destination path '%s' is not a directory." % dstpath)

    # Mutually Exclusive
    if OVERWRITE and SKIP:
        error(
            "Options --overwrite and --skip are mutually exlusive. (use one or the other, not both)"
        )

    # Parse the source file
    print_on_true("Parsing %s..." % sourcefile, VERBOSE)
    try:
        parsed_tree = Parser.parse_file(sourcefile)
    except ParseError as e:
        error(e)

    # Print the graphical tree and Exit.
    if PRINT_TREE:
        print_tree(parsed_tree)
        sys.exit(0)

    # Create paths from tree nodes
    print_on_true("Creating tree paths", VERBOSE)
    paths: Dict[str, List[str]] = Normalizer.normalize(parsed_tree, dstpath)

    # If Overwrite and Skip both are false
    print_on_true("Checking existing tree paths", VERBOSE)
    if not OVERWRITE and not SKIP:
        if count := print_existing_paths(paths["files"]):
            error(
                f"\nFix {count} issue{'s' if count > 1 else ''} before moving forward."
            )

    # Create the files and dirs finally
    print_on_true("Creating tree on filesystem", VERBOSE)
    creation_count: Tuple[int, int] = TreeBuilder.build(
        paths, skip=SKIP, verbose=VERBOSE
    )
    print_on_true("Done.\n", VERBOSE)

    # Completion message
    print(
        "%d directories and %d files have been created."
        % (creation_count[0], creation_count[1])
    )


def parse_args():
    """Parse command-line arguments and return."""

    parser = ArgumentParser(
        prog=PROGRAM,
        usage="%(prog)s [OPTIONS]",
        epilog="%s %s" % (PROGRAM.title(), VERSION),
        description="A CLI tool to create directory structures from a structure file.",
    )

    parser.add_argument("src", help="source file (with .tree extension)")
    parser.add_argument(
        "dst",
        nargs="?",
        default=".",
        help="where to create the tree structure (default: %(default)s)",
    )
    parser.add_argument(
        "-g",
        "--graphical",
        action="store_true",
        help="show source file as graphical tree and exit",
    )
    parser.add_argument(
        "-o", "--overwrite", action="store_true", help="overwrite existing files"
    )
    parser.add_argument("-s", "--skip", action="store_true", help="skip existing files")
    # parser.add_argument(
    #     "-v", "--version", action="version", version="%s %s" % (PROGRAM, VERSION)
    # )
    parser.add_argument(
        "-v", "--verbose", action="store_true", help="increase verbosity"
    )

    return parser.parse_args()


def error(message: str):
    """Print `message` and exit with status `1`. Use upon errors only."""
    print(message)
    sys.exit(1)


def print_existing_paths(paths: List[str]) -> int:
    """Print existing paths. Return the number of paths that exist."""
    count = 0
    if existing_paths := get_existing_paths(paths):
        count = len(existing_paths)
        for path in existing_paths:
            print("Warning: File '%s' already exists" % path)

    return count



"""  Responsible for reading and parsing the structure file (in `.tree` format),
that users provide to define the directory structure. """

from maketree.utils import is_valid_dir, is_valid_file
from typing import List


class ParseError(Exception):
    def __init__(self, *args: object) -> None:
        super().__init__(*args)
        self.args = args


class Parser:

    @classmethod
    def parse_file(cls, filepath: str):
        """Parse `filepath` .tree file and return the tree in a usable format (e.g, `dict` or `list`)"""
        with open(filepath, encoding="utf-8") as srcfile:
            raw_data = srcfile.readlines()

        return Parser._parse_lines(raw_data)

    @classmethod
    def _parse_lines(cls, lines: List[str]):
        """Parse `lines` into tree structure"""

        stack = []  # Keep track of parent dirs
        tree = []  # Final parsed tree (list of dicts)

        for i, line in enumerate(lines):
            line = line.rstrip()

            # Empty line?
            if not line:
                continue

            # Indentation level of current entry
            indent_level: int = (len(line) - len(line.lstrip())) // 4

            if line.endswith("/"):  # Its a Directory
                item = {
                    "type": "directory",
                    "name": line.strip().rstrip("/"),
                    "children": [],
                }

                # Validate dir name
                valid = is_valid_dir(item["name"])
                if valid is not True:
                    raise ParseError(
                        "at line %d, in '%s', %s" % (i + 1, item["name"], valid)
                    )

                # Pop from stack til the correct parent
                while stack and stack[-1]["indent_level"] >= indent_level:
                    stack.pop()

                if stack:
                    # Add this dir to its parent's children
                    stack[-1]["item"]["children"].append(item)
                else:
                    # Top Level directory, stack is empty
                    tree.append(item)

                # Push this dir onto stack
                stack.append({"item": item, "indent_level": indent_level})

            else:  # Its a File
                item = {"type": "file", "name": line.strip()}

                # Validate file name
                valid = is_valid_file(item["name"])
                if valid is not True:
                    raise ParseError(
                        "at line %d, in '%s', %s" % (i + 1, item["name"], valid)
                    )

                # Pop from stack til the correct parent
                while stack and stack[-1]["indent_level"] >= indent_level:
                    stack.pop()

                if stack:
                    # Add this file to current parent's children
                    stack[-1]["item"]["children"].append(item)
                else:
                    # Top Level file, stack empty
                    tree.append(item)

        return tree



""" Contains Helper code to keep core logic clean. (things that don't fit anywhere, fit here) """

from os.path import exists, splitext
from platform import system
from typing import List, Dict, Union


# Windows, Darwin, Linux
OS: str = system()


def get_nonexisting_paths(paths: List[str]) -> List[str]:
    """Returns a list of non-existing paths from `paths` list."""
    return list(filter(lambda p: not exists(p), paths))


def get_existing_paths(paths: List[str]) -> List[str]:
    """Returns a list of existing paths from `paths` list."""
    return list(filter(lambda p: exists(p), paths))


def is_valid_extension(extension: str) -> bool:
    """
    ### Is Valid Extension
    Returns `True` if extension is valid, `False` otherwise.
    `extension` must contain a period `.`

    An extension is valid if it follows below criteria:
    - extension must be non-empty (excluding period `.`)
    - extension must have a period `.`
    - extension must not contain symbols, whitespaces
        - `\\/:*?"<>|` are illegal on Windows
        - `/:` are illegal on Mac
        - `/` are illegal on Linux
    """
    if not extension:
        return False

    if len(extension) < 2:
        return False

    if not extension.startswith("."):
        return False

    if extension.count(".") > 1:
        return False

    if OS == "Windows":
        # Got Illegals?
        if contains_chars(extension, r' \/:*?"<>|'):
            return False

    elif OS == "Darwin" and contains_chars(extension, "/:"):
        return False

    elif OS == "Linux" and "/" in extension:
        return False

    return True


def is_valid_file(filename: str) -> Union[bool, str]:
    """
    ### Is Valid File
    Validates filename. Returns `True` if valid, Returns `str` if invalid.
    This `str` is the cause of filename invalidation.

    #### ARGS:
    - `filename`: name of the file

    #### Note:
    This function is not a stripped down version of itself. (specific to needs of the `Parser`, minimal but fast)
    """
    if not filename:
        return "file name must not be empty"

    # Split filepath into root and extension
    root, ext_ = splitext(filename)

    # Root must not be empty
    if not root:
        return "invalid file name"

    # Check for illegal chars
    if OS == "Windows":
        if contains_chars(root, r'\/:?*<>"|'):
            return 'illegal characters are not allowed: \\/:?*<>|"'
    elif OS == "Darwin":
        if contains_chars(root, r"/:<>"):
            return "illegal characters are not allowed: /:?<>"
    else:  # Linux
        if contains_chars(root, r"/:<>"):
            return "illegal characters are not allowed: /:?<>"

    if ext_ and not is_valid_extension(ext_):
        return "invalid file extension"

    return True


def is_valid_dir(dirname: str) -> Union[bool, str]:
    """
    ### Is Valid Dirpath
    Validates directory name. Returns `True` if valid, Returns `str` if invalid.
    This `str` contains the reason for dir being invalid.

    #### ARGS:
    - `dirname`: the path to validate

    #### Note:
    This function is not a stripped down version of itself. (specific to needs of the `Parser`, minimal but fast)
    """
    if not dirname:
        return "directory must not be empty."

    # Check for illegal chars
    if OS == "Windows":
        if contains_chars(dirname, r'\/:?*<>"|'):
            return 'illegal characters are not allowed: \\/:?*<>|"'
    elif OS == "Darwin":
        if contains_chars(dirname, r"/:<>"):
            return "illegal characters are not allowed: /:?<>"
    else:
        if contains_chars(dirname, r"/:<>"):
            return "illegal characters are not allowed: /:?<>"

    return True


def contains_chars(string: str, chars: str) -> bool:
    """
    ### Contains
    Checks whether `string` contains a character from `chars`.
    Returns `True` if it does, `False` if does not.
    """
    return any(char for char in chars if char in string)


def print_on_true(string: str, bool_: bool):
    """
    Prints the string to terminal only if the `bool_` is true.
    _(Never thought i'd ever create/need such function, but it is useful for
    printing verbose messages dynamically)_

    **Use it only for Verbose messages.**
    """
    if bool_:
        print(string)


def print_tree(tree: List[Dict]):
    """Prints the parsed `tree` in a graphical format. _(Not perfect but, gets the job done)_"""
    tab = 0
    BAR = "│   "
    LINK = "├───"
    LINK_LAST = "└───"
    FMT_STR = f"%s%s %s"

    def traverse(node: Dict, childs: int):
        nonlocal tab
        count = 0  # keeps track of child counts

        for child in node.get("children", []):
            count += 1

            if count == childs:
                # Last Child
                print(FMT_STR % (BAR * tab, LINK_LAST, child["name"]))
            else:
                # Others
                print(FMT_STR % (BAR * tab, LINK, child["name"]))

            if child["type"] == "directory" and child["children"]:
                tab += 1
                traverse(child, len(child["children"]))
        tab -= 1
        return

    # Root dot.
    print(".")

    traverse(
        node={
            "type": "directory",
            "name": ".",
            "children": tree,
        },
        childs=len(tree),
    )



""" Entry point of the project (to run directly, without installing) """

import os
import sys


if __package__ == "":
    path = os.path.dirname(os.path.dirname(__file__))
    sys.path.insert(0, path)

from maketree import cli

if __name__ == "__main__":
    cli.main()
